experiment = setRefClass("experiment",
  fields = list(
    fields="list",
    nuclei="list",
	  selections="list",
	  metaData="list",
    is.fetched = "logical"
  ),
  methods = list(
    show = function(){
      print(content['name'])
    },
    setPopulation = function(){
      population <<- "experiments"
    },
    setModel = function(){
      model <<- "experiment"
    },
    postInitialize = function(){
      is.fetched <<- FALSE
    },
    fetchData = function(){
      if(!is.fetched){
        .self$setNuclei()
        .self$setFields()
        .self$setSelections()
        .self$setMetaData()
        is.fetched <<- TRUE
      }
    },
    getStructures = function(){
      return(.self$getContent("structures"))
    },
    getVirtualStructures = function(){
      l = .self$getContent("virtualStructures")
      if(is.null(l) || length(l)==0) return(list())
      else{
        offset = length(.self$getStructures())
        names(l)=offset:(offset+length(l)-1)
        return(l)
      }
    },
    getStructuresByName = function(){
      structures=list()
      for(structure in .self$getStructures()){
          structures[[structure$name]] = structure
      }
      return(structures)
    },
    getStructuresNames = function(){
      return(names(.self$getStructuresByName()))
    },
#     getStructure = function(tag){
#       
#     },
#     getStructures = function(tags){
#       
#     },
    getVirtualStructuresByName = function(){
      virtualstructures=list()
      for(virtualstructure in .self$getVirtualStructures()){
        virtualstructures[[virtualstructure$name]] = virtualstructure
      }
      return(virtualstructures)
    },
    getVirtualStructuresNames = function(){
      return(names(.self$getVirtualStructuresByName))
    },
    getAllStructures = function(){
      return(c(.self$getStructures(),.self$getVirtualStructures()))
    },
	  setMetaData = function(data=NULL){
		  if(!is.null(data)){
			  for(name in names(data)){
				  metaData[[name]] <<- as.integer(data[[name]])
			  }
		  }
		  else{
			  s = .self$getAllStructures()
			  for(idx in names(s)){
				  metaData[[s[[idx]]$name]] <<- as.integer(idx)
			  }
		  }
	  },
    extractObjectIndexes = function(df, columnNames){ 
      ObjectIndexesByStructureIndex = list()
      for (name in names(columnNames)){
        structureIndex = columnNames[[name]]
        structureColumnValues = na.omit(unique(df[[name]])) 
        ObjectIndexesByStructureIndex[[structureIndex]] = unique(c(ObjectIndexesByStructureIndex[[structureIndex]],structureColumnValues))
      }
      return(ObjectIndexesByStructureIndex)
    },
    extractSelection = function(df,selectionName='currentSelection',description=NULL, columnNames=NULL,save=TRUE){
      if (is.null(columnNames)) {
        for (column.name in names(df)){
          structureName = strsplit(column.name,'.idx')[[1]]
          if(nchar(structureName)<nchar(column.name)){ # si le nom est plus court c'est qu'il a trouve .idx donc c'est un nom de colonne 
            structureIndex = toString(metaData[[structureName]])
            columnNames[[column.name]]=structureIndex
          }
        }
      } else { # correction au cas ou l'utilisateur ai donne des indices numeriques et non des strings
        for (column.name in names(columnNames)) {
          columnNames[[column.name]]=toString(columnNames[[column.name]])
        }
      }
      l = list()
      splitDfByNuclei = split(df, f = df$nucleus.id)
      for(nucleusId in names(splitDfByNuclei)){
        l[[nucleusId]] = .self$extractObjectIndexes(splitDfByNuclei[[nucleusId]], columnNames)
      }
      c = list(experiment_id=oid,name=selectionName,nuclei=l)
      sel=selection$new(connector=connector,content=c,parents=list(experiment = .self),db=db)
      sel$initFields()
      if(save){
        sel$save()
      }
      return(sel)
    },
    getAllStructuresNames = function(){
      return(c(.self$getStructuresNames(),.self$getVirtualStructuresNames()))
    },
    getTags = function(filter=NULL){
      if (is.null(filter)) {
        l = names(.self$nuclei)
      } else {
        l = filter
      }
      tags=numeric(0)
      for (id in l) {
        t=.self$nuclei[[id]]$getContent("tag")   
        if (!is.null(t)) tags[id]=t
        else  tags[id]=0
      }
      return(tags)
    },
    setNuclei = function(){
      ns = .self$getRelateds("nuclei")
      nuclei <<- ns$elements
    },
	  setSelections = function(){
		  s = .self$getRelateds("selections")
		  selections <<- s$indexBy("name")
	  },
    createWholeSelection = function(){
      l = list(experiment_id = oid, name = 'Whole',description='',nuclei = list())
      c = .self$getObjectsCount()
      nuclei.ids = names(nuclei)
      for(nucleus.id in nuclei.ids){
        l[['nuclei']][[nucleus.id]] = list()
        nucleus.structures.indexes = names(.self$getStructures())
        for(structure.index in nucleus.structures.indexes){
          nucleus = .self$nuclei[[nucleus.id]]
          nucleus$objects3dCount[[structure.index+1]]
          l[['nuclei']][[nucleus.id]][[structure.index]] = c(1:nucleus$objects3dCount[[structure.index+1]])
        }
      }
      #TODO construction de l'objet selection associé à partir des indexes
      return(selection$new(connector = connector,content = l,parents = parents,db = db))
    },
    createEmptySelection = function(){
      return(selection$new(connector = connector,content = list(experiment_id = oid, name = 'Empty',description='',nuclei = list()),parents = parents,db = db))
    },
    getNucleiFieldsInformation = function(){
      l = list()
      for(nucleus_id in names(exp$nuclei)){
        l[[nucleus_id]] = exp$nuclei[[nucleus_id]]$getFieldNameAndIdx()
      }
      return(l)
    },
    setFields = function(){
      fs = .self$getRelateds("fields")
      fields <<- fs$elements
    },
    getKey = function(structures){
      if (length(structures)==1) structures=c(structures, structures)
      key = structures[1]
      if (length(structures>1)) {
        for (i in 2:length(structures)) {
          key=paste(key, i,  sep=";")
        }
      }
      return(key)
    },
    extractStructureData = function(structures,keys,filter=NULL,addIdx=FALSE,useMetaData=TRUE,annotations=TRUE){
      if(!is.fetched) .self$fetchData()
      list.data=list()
      if (is.null(filter)) {
        l = names(.self$nuclei)
      } else {
        l = filter
      }
      for (nucleus_id in l) {
        list.data[[nucleus_id]] = .self$nuclei[[nucleus_id]]$extractStructureData(structures, keys, addIdx,useMetaData,annotations)
      }
      data=rbind.fill(list.data)
      if (!is.null(data) && nrow(data)>0 && annotations) {
        data$experiment.name = .self$getContent("name")
        data$project.name = db
        data$host.name = .self$getHost()
      }
      if (nrow(data)>0 && any(is.na(data))) warning('Missing values in dataframe: either missing keys or wrong number of object')
      return(data)
    },
    extractObjectData = function(structure,keys,filter=NULL,addIdx=FALSE,useMetaData=TRUE,annotations=TRUE){
      if(!is.fetched) .self$fetchData()
      list.data=list()
      if (is.null(filter)) {
        l = names(.self$nuclei)
      } else {
        l = filter
      }
      for (nucleus_id in l) {
        list.data[[nucleus_id]]=.self$nuclei[[nucleus_id]]$extractObjectData(structure, keys, addIdx, useMetaData,annotations)
      }
      data=rbind.fill(list.data)
      if (!is.null(data) && nrow(data)>0 && annotations) {
        data$experiment.name = .self$getContent("name")
        data$project.name = db
        data$host.name = .self$getHost()
      }
      if (is.null(data)) warning('Empty dataframe. You might want to choose different structures or different keys.')
      if (nrow(data)>0 && any(is.na(data))) warning('Missing values in dataframe: either missing keys or wrong number of object')
      return(data)
    },
    getObjectsCount = function(structure,filter=NULL){
      if (is.null(filter)) {
        l = names(.self$nuclei)
      } else {
        l = filter
      }
      countsList=numeric(0)
      for (nucleus_id in l) {
        nucleus = .self$nuclei[[nucleus_id]]
        countsList[nucleus_id] = nucleus$objects3dCount[[structure+1]]
      }
      return(countsList)
    },
    getTagsAndStructures = function(structures=NULL){
      if(!is.fetched) .self$fetchData()
      df = .self$getTags()
      structNames=.self$getAllStructuresNames()
      for (i in 1:length(structNames)) {
        df=cbind(df, .self$getObjectsCount(i-1))
      }
      df=data.frame(df)
      names(df)=c("tags", structNames)
      return(df)
    },
    getFilterFromTags = function(tags){
      if(is.null(tags)) return(names(nuclei))
      else{
        filter = c()
        info =.self$getTags()
        for(id in names(info)){
          if(info[id] %in% tags) filter = c(filter,id)
        }
        return(filter)
      }
    },
    getFilterFromSelections = function(selections,intersection=FALSE){
      if(is.null(selections)) return(names(nuclei))
      else{
        if(intersection){
          filter = names(nuclei)
          for(selectionName in names(selections)){
            selection = .self$selections[[selectionName]]
            filter = intersect(filter,selection$getFilter())
          }
          return(filter)
        }else{
         filter = c()
         for(selectionName in selections){
           selection = .self$selections[[selectionName]]
           filter = union(filter,selection$getFilter())
         }
         return(filter)
        }
      }
    },
    getFilterFromSelectionsAndTags = function(selections,tags){
      filter1 = .self$getFilterFromSelections(selections)
      filter2 = .self$getFilterFromTags(tags)
      return(intersect(filter1,filter2))
    },
    getFilterFromSelectionsTagsAndObjectNumbers = function(selections, tags, objectNumberPerStructure){
      filter1 = .self$getFilterFromSelections(selections)
      filter2 = .self$getFilterFromTags(tags)
      filter3 = .self$getObjectNumberFilter(objectNumberPerStructure)
      return(intersect(intersect(filter1,filter2),filter3))
    },
    getFilter = function(s1=-1,n1,s2=-1,n2,tags){
      if(!is.fetched) .self$fetchData()
        names=.self$getStructuresNames()
        nuc.info=.self$getInfo()
        if (s1>=0 & s2>=0) filter=row.names(
          nuc.info[nuc.info[,
					names[s1+1]]%in%n1 & nuc.info[, 
          names[s2+1]]%in%n2 & nuc.info[,
					"tags"]%in%tags, ]
        )
        else if (s1>=0) filter=row.names(
          nuc.info[nuc.info[,
          names[s1+1]]%in%n1 & nuc.info[, "tags"]%in%tags, ]
        )
        else if (s2>=0) filter=row.names(nuc.info[nuc.info[,
          names[s2+1]]%in%n2 & nuc.info[,
          "tags"]%in%tags, ]
        )
        else filter=row.names(nuc.info[nuc.info[, "tags"]%in%tags, ])
      return(filter)
	  },
    splitDataFrameBySelection = function(dataframe,selection){
      
    },
    appendECDF=function(df, measure, split=NULL, xlim=NULL, subtract.unif=FALSE) {
      if (subtract.unif) {
        if (is.null(xlim)) func = function(x) {return(ecdf(x)(x)-punif(min=min(x), max=max(x), x))}
        else func = function(x) {return(ecdf(x)(x)-punif(min=xlim[1], max=xlim[2], x))}
      }
      else func = function(x) {return(ecdf(x)(x))}
      yname = paste(measure, "ecdf", sep=".")
      if (is.null(split)) {
        df$newColumnECDF = func(df[, measure])
      } else {
        df = ddply(df, c(split), .fun=function(x, colName) transform(x, newColumnECDF=func(x[, colName])), colName=measure)
      }
      idx = which(names(df)=="newColumnECDF")
      colnames(df)[idx]=yname
      
      if (!is.null(split)&length(split)==1) { ## adds number of entries
        df = ddply(df, c(split), .fun=function(x, split.name) transform(x, ecdf.split.colname=paste(x[1, split.name], " (N=", nrow(x),")", sep="")), split.name=split)
        color.name=paste(split, "ecdf", sep=".")
        names(df)[which(names(df)=="ecdf.split.colname")] = color.name
      }
      if (subtract.unif) { #summary
        
        print(summaryBy(data=df, as.formula(paste(yname, paste0(split, collapse="+"), sep="~")), FUN=c(max, min, length)))
      }
      return(df)
    },
    plotECDF=function(df, measure, color=NULL, linetype=NULL, size=0.5, xlim=NULL, facet.wrap=NULL) {
      require(ggplot2)
      color.ecdf = paste(color, "ecdf", sep=".")
      if (!is.null(xlim)&&length(xlim)!=2) xlim=c(min(df[,measure]), max(df[,measure]))
      yvar=paste(measure, "ecdf", sep=".")
      split.vars = c(color, linetype, facet.wrap)
      if (length(split.vars)>1|(length(split.vars==1)&is.null(color))) {
        dfPlot = appendECDF(df, measure, split.vars, xlim, FALSE)
        if (is.null(color)) p = ggplot(dfPlot) + geom_step(aes_string(x=measure, y=yvar, linetype=linetype), color=2, size=size) + ylab("")
        else p = ggplot(dfPlot) + geom_step(aes_string(x=measure, y=yvar, color=color, linetype=linetype), size=size) + ylab("")
        p = p+ scale_y_continuous(limits=c(0, 1))
        if (!is.null(facet.wrap)) p = p+  facet_wrap(as.formula(paste("~", facet.wrap, sep="")))
        return(p)
      } else if (!is.null(color)) { ## juste les couleurs -> on ajoute le nombre d'entrées
        dfPlot = appendECDF(df, measure, color, xlim, FALSE)
        p = ggplot(dfPlot) + geom_step(aes_string(x=measure, y=yvar, color=color), size=size) + ylab("")
        p = p+ scale_y_continuous(limits=c(0, 1))
        if (!is.null(facet.wrap)) p = p+  facet_wrap(as.formula(paste("~", facet.wrap, sep="")))
        return(p)
      } else {
        dfPlot = appendECDF(df, measure, facet.wrap, xlim, FALSE)
        p=ggplot(dfPlot) + geom_step(aes_string(x=measure, y=yvar), color=2, size=size)
        p = p+ scale_y_continuous(limits=c(0, 1))
        return(p)
      }
    },
    getObjectNumberFilter = function(structure.object.number.list) {
      nuc.info=.self$getTagsAndStructures()
      filter = row.names(nuc.info)
      for (structure.name in names(structure.object.number.list)) {
        authorized.numbers = structure.object.number.list[[structure.name]]
        authorized.rows = nuc.info[nuc.info[[structure.name]] %in% authorized.numbers,]
        filter=intersect(filter, row.names(authorized.rows))
      }
      return(filter)
    },
    appendSelections = function(df, selection.names, column.name="population", default.name="no population") {
      require(plyr)
      appendPopulationName = function(df, column.name, population.name, default.name, sep=".") {
        if (df[1,column.name]==default.name) df[1,column.name]=population.name
        else df[1,column.name]=paste(df[1,column.name], population.name, sep=sep)
        return(df)
      }
      df[,column.name] = default.name
      for (selection in selection.names) {
        if (!is.null(selections[[selection]])) {
          lines = df$nucleus.id%in%selections[[selection]]$getFilter()
          df[lines,]=adply(df[lines,], 1, .fun=appendPopulationName, column.name=column.name, population.name=selection, default.name=default.name)
        } else print(paste("selection:", selection, " not found"))
      }
      return(df)
    }
  ),
  contains="mongoDbPersistentObject"
)




